@@ -2,6 +2,8 @@
 #
 # reporting-sync-state
 #
+# Copyright (C) 2018 Chris Lamb <lamby@debian.org>
+#
 # This program is free software.  It is distributed under the terms of
 # the GNU General Public License as published by the Free Software
 # Foundation; either version 2 of the License, or (at your option) any
@@ -14,12 +16,15 @@
 #
 # You should have received a copy of the GNU General Public License
 # along with this program.  If not, you can find it on the World Wide
-# Web at http://www.gnu.org/copyleft/gpl.html, or write to the Free
+# Web at https://www.gnu.org/copyleft/gpl.html, or write to the Free
 # Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
 # MA 02110-1301, USA.
 
-use strict;
+package reporting_sync_state;
+
+use v5.20;
 use warnings;
+use utf8;
 use autodie;
 
 use Getopt::Long();
@@ -27,18 +32,18 @@ use File::Basename qw(basename);
 use YAML::XS ();
 use MIME::Base64 qw(encode_base64);
 
-use Lintian::Lab;
+use Lintian::Deb822::Parser qw(visit_dpkg_paragraph_string);
 use Lintian::Relation::Version qw(versions_comparator);
-use Lintian::Util qw(
+use Lintian::Reporting::Util qw(
   find_backlog
   load_state_cache
-  open_gz
   save_state_cache
-  strip
-  visit_dpkg_paragraph
+);
+use Lintian::Util qw(
+  open_gz
 );
 
-my $DEFAULT_CHECKSUM = 'sha256';
+my $DEFAULT_CHECKSUM = 'Sha256';
 my (%KNOWN_MEMBERS, %ACTIVE_GROUPS);
 my $CONFIG;
 my %OPT;
@@ -88,8 +93,24 @@ sub required_cfg_non_empty_list_value {
     return $v;
 }
 
+sub optional_cfg_list_value {
+    my (@keys) = @_;
+    my $v = $CONFIG;
+    for my $key (@keys) {
+        if (not exists($v->{$key})) {
+            return [];
+        }
+        $v = $v->{$key};
+    }
+    if (ref($v) ne 'ARRAY') {
+        my $k = join('.', @keys);
+        die("Invalid configuration: ${k} must be a list (or missing)\n");
+    }
+    return $v;
+}
+
 sub main {
-    my ($state_dir, $state, $archives);
+    my ($state_dir, $state, $archives, %blacklist);
     STDOUT->autoflush;
     Getopt::Long::config('bundling', 'no_getopt_compat', 'no_auto_abbrev');
     Getopt::Long::GetOptions(%OPT_HASH) or die("error parsing options\n");
@@ -105,6 +126,7 @@ sub main {
           . (scalar(keys(%{$state->{'groups'}})))
           . ' groups');
     $archives = required_cfg_value('archives');
+    %blacklist = map { $_ => 1 } @{optional_cfg_list_value('blacklist')};
     for my $archive (sort(keys(%{$archives}))) {
         log_debug("Processing archive $archive");
         my $path = required_cfg_value('archives', $archive, 'base-dir');
@@ -117,7 +139,7 @@ sub main {
           = required_cfg_non_empty_list_value('archives', $archive,
             'distributions');
         local_mirror_manifests($state, $path, $distributions, $components,
-            $archs);
+            $archs, \%blacklist);
     }
 
     cleanup_state($state);
@@ -208,19 +230,20 @@ sub add_member_to_group {
     $KNOWN_MEMBERS{"${group_id} ${member_id}"} = 1;
     $ACTIVE_GROUPS{$group_id} = 1;
 
-    if (!exists($member->{$DEFAULT_CHECKSUM})
-        || $member->{$DEFAULT_CHECKSUM} ne $member_data->{$DEFAULT_CHECKSUM}) {
-        if (exists($member->{$DEFAULT_CHECKSUM})) {
+    if (!exists($member->{lc $DEFAULT_CHECKSUM})
+        || $member->{lc $DEFAULT_CHECKSUM} ne
+        $member_data->{lc $DEFAULT_CHECKSUM}) {
+        if (exists($member->{lc $DEFAULT_CHECKSUM})) {
             # This seems worth a note even if the group is already out of date
+            my $lowercase = lc $DEFAULT_CHECKSUM;
             log_debug(
                 "${group_id} is out-of-date: ${member_id} checksum mismatch"
-                  . " ($member->{$DEFAULT_CHECKSUM} != $member_data->{$DEFAULT_CHECKSUM})"
-            );
+                  . " ($member->{$lowercase} != $member_data->{$lowercase})");
         } elsif (not $group_data->{'out-of-date'} and not $new_group) {
             log_debug("${group_id} is out-of-date: New member (${member_id})");
         }
         $group_data->{'out-of-date'} = 1;
-        $member->{$DEFAULT_CHECKSUM} = $member_data->{$DEFAULT_CHECKSUM};
+        $member->{lc $DEFAULT_CHECKSUM} = $member_data->{lc $DEFAULT_CHECKSUM};
     }
     delete($member->{'sha1'});
 
@@ -306,7 +329,7 @@ sub cleanup_group_state {
         exists($group_data->{'processing-errors'})
         and (not exists($group_data->{'last-error-by'})
             or $group_data->{'last-error-by'} ne $OPT{'desired-version'})
-      ) {
+    ) {
         log_debug(
             "Clearing error-counter for ${group_id}: New version of lintian");
         delete($group_data->{'processing-errors'});
@@ -333,23 +356,30 @@ sub cleanup_group_state {
 
 # Helper for local_mirror_manifests - it parses a paragraph from Sources file
 sub _parse_srcs_pg {
-    my ($state, $extra_metadata, $paragraph) = @_;
-    my $dir = $paragraph->{'directory'}//'';
-    my $group_id = $paragraph->{'package'} . '/' . $paragraph->{'version'};
+    my ($state, $blacklist, $extra_metadata, $paragraph) = @_;
+    my $dir = $paragraph->{'Directory'}//'';
+    my $group_id = $paragraph->{'Package'} . '/' . $paragraph->{'Version'};
     my $member_id = "source:${group_id}";
     my (%data, %group_metadata, $group_mirror_md);
+    if (exists $blacklist->{$paragraph->{'Package'}}) {
+        log_debug("Ignoring blacklisted package src:$paragraph->{'Package'}");
+        return;
+    }
     # only include the source if it has any binaries to be checked.
     # - Otherwise we may end up checking a source with no binaries
     #   (happens if the architecture is "behind" in building)
     return unless $ACTIVE_GROUPS{$group_id};
     $dir .= '/' if $dir;
-    foreach my $f (split m/\n/, $paragraph->{"checksums-${DEFAULT_CHECKSUM}"}){
-        strip($f);
+    foreach my $f (split m/\n/, $paragraph->{"Checksums-${DEFAULT_CHECKSUM}"}){
+
+        # trim both ends
+        $f =~ s/^\s+|\s+$//g;
+
         next unless $f && $f =~ m/\.dsc$/;
         my ($checksum, undef, $basename) = split(m/\s++/, $f);
         my $b64_checksum = encode_base64(pack('H*', $checksum));
         # $dir should end with a slash if it is non-empty.
-        $data{$DEFAULT_CHECKSUM} = $b64_checksum;
+        $data{lc $DEFAULT_CHECKSUM} = $b64_checksum;
         $data{'path'} = $extra_metadata->{'mirror-dir'}  . "/$dir" . $basename;
         last;
     }
@@ -369,30 +399,36 @@ sub _parse_srcs_pg {
 
 # Helper for local_mirror_manifests - it parses a paragraph from Packages file
 sub _parse_pkgs_pg {
-    my ($state, $extra_metadata, $type, $paragraph) = @_;
+    my ($state, $blacklist, $extra_metadata, $type, $paragraph) = @_;
     my ($group_id, $member_id, %data, %group_metadata, $b64_checksum);
-    my $package = $paragraph->{'package'};
-    my $version = $paragraph->{'version'};
-    my $architecture = $paragraph->{'architecture'};
-    if (not defined($paragraph->{'source'})) {
-        $paragraph->{'source'} = $package;
-    } elsif ($paragraph->{'source'} =~ /^([-+\.\w]+)\s+\((.+)\)$/) {
-        $paragraph->{'source'} = $1;
-        $paragraph->{'source-version'} = $2;
+    my $package = $paragraph->{'Package'};
+    my $version = $paragraph->{'Version'};
+    my $architecture = $paragraph->{'Architecture'};
+    if (not defined($paragraph->{'Source'})) {
+        $paragraph->{'Source'} = $package;
+    } elsif ($paragraph->{'Source'} =~ /^([-+\.\w]+)\s+\((.+)\)$/) {
+        $paragraph->{'Source'} = $1;
+        $paragraph->{'Source-version'} = $2;
     }
-    if (not defined($paragraph->{'source-version'})) {
-        $paragraph->{'source-version'} = $paragraph->{'version'};
+    if (exists $blacklist->{$paragraph->{'Source'}}) {
+        log_debug("Ignoring binary package $package: it is part of "
+              . "blacklisted source package $paragraph->{'Source'}");
+        return;
     }
-    $group_id = $paragraph->{'source'} . '/' . $paragraph->{'source-version'};
+    if (not defined($paragraph->{'Source-Version'})) {
+        $paragraph->{'Source-Version'} = $paragraph->{'Version'};
+    }
+    $group_id = $paragraph->{'Source'} . '/' . $paragraph->{'Source-Version'};
     $member_id = "${type}:${package}/${version}/${architecture}";
     $data{'path'}
       = $extra_metadata->{'mirror-dir'} . '/' . $paragraph->{'filename'};
-    $b64_checksum = encode_base64(pack('H*', $paragraph->{$DEFAULT_CHECKSUM}));
-    $data{$DEFAULT_CHECKSUM} = $b64_checksum;
+    $b64_checksum
+      = encode_base64(pack('H*', $paragraph->{lc $DEFAULT_CHECKSUM}));
+    $data{lc $DEFAULT_CHECKSUM} = $b64_checksum;
 
     $group_metadata{'mirror-metadata'}{'maintainer'}
-      = $paragraph->{'maintainer'};
-    if (my $uploaders = $paragraph->{'uploaders'}) {
+      = $paragraph->{'Maintainer'};
+    if (my $uploaders = $paragraph->{'Uploaders'}) {
         my @ulist = split(/>\K\s*,\s*/, $uploaders);
         $group_metadata{'mirror-metadata'}{'uploaders'} = \@ulist;
     }
@@ -413,43 +449,60 @@ sub _parse_pkgs_pg {
 # $dists  - listref of dists to consider (e.g. ['unstable'])
 # $components  - listref of components to consider (e.g. ['main', 'contrib', 'non-free'])
 # $archs  - listref of archs to consider (e.g. ['i386', 'amd64'])
+# $blacklist  - hashref of source packages to ignore (e.g. {'gcc-8-cross-ports' => 1})
 #
 sub local_mirror_manifests {
-    my ($state, $mirdir, $dists, $components, $archs) = @_;
+    my ($state, $mirdir, $dists, $components, $archs, $blacklist) = @_;
     foreach my $dist (@$dists) {
         foreach my $component (@{$components}) {
             my $srcs = "$mirdir/dists/$dist/$component/source/Sources";
-            my ($srcfd, $srcsub);
+
             my %extra_metadata = (
                 'component' => $component,
                 'mirror-dir' => $mirdir,
             );
+
             # Binaries have a "per arch" file.
             # - we check those first and then include the source packages that
             #   are referred to by these binaries.
+
             my $dist_path = "$mirdir/dists/$dist/$component";
-            foreach my $arch (@{$archs}) {
+            for my $arch (@{$archs}) {
+
                 my $pkgs = "${dist_path}/binary-$arch/Packages";
+                my $pkgfd = _open_data_file($pkgs);
+                local $/ = undef;
+                my $pkgstring = <$pkgfd>;
+                close $pkgfd;
+
+                my $binsub = sub {
+                    _parse_pkgs_pg($state, $blacklist, \%extra_metadata,
+                        'binary', @_);
+                };
+                visit_dpkg_paragraph_string($binsub, $pkgstring);
+
                 my $upkgs
                   = "${dist_path}/debian-installer/binary-$arch/Packages";
-                my $pkgfd = _open_data_file($pkgs);
-                my $binsub = sub {
-                    _parse_pkgs_pg($state, \%extra_metadata, 'binary', @_);
-                };
-                my $upkgfd;
+                my $upkgfd = _open_data_file($upkgs);
+                local $/ = undef;
+                my $upkgstring = <$upkgfd>;
+                close $upkgfd;
+
                 my $udebsub = sub {
-                    _parse_pkgs_pg($state, \%extra_metadata, 'udeb', @_);
+                    _parse_pkgs_pg($state, $blacklist, \%extra_metadata,
+                        'udeb', @_);
                 };
-                visit_dpkg_paragraph($binsub, $pkgfd);
-                close($pkgfd);
-                $upkgfd = _open_data_file($upkgs);
-                visit_dpkg_paragraph($udebsub, $upkgfd);
-                close($upkgfd);
+                visit_dpkg_paragraph_string($udebsub, $upkgstring);
             }
-            $srcfd = _open_data_file($srcs);
-            $srcsub = sub { _parse_srcs_pg($state, \%extra_metadata, @_) };
-            visit_dpkg_paragraph($srcsub, $srcfd);
-            close($srcfd);
+
+            my $srcfd = _open_data_file($srcs);
+            local $/ = undef;
+            my $srcstring = <$srcfd>;
+            close $srcfd;
+
+            my $srcsub
+              = sub { _parse_srcs_pg($state, $blacklist, \%extra_metadata, @_) };
+            visit_dpkg_paragraph_string($srcsub, $srcstring);
         }
     }
     return;
@@ -491,7 +544,7 @@ sub log_debug {
 
 sub usage {
     my $cmd = basename($0);
-    my $me = "dplint $cmd";
+    my $me = "lintian $cmd";
     print <<EOF;
 Internal command for the Lintian reporting framework
 Usage: $me <args>
@@ -510,6 +563,8 @@ EOF
     exit(0);
 }
 
+1;
+
 # Local Variables:
 # indent-tabs-mode: nil
 # cperl-indent-level: 4
